import re
import urllib.parse
import w3lib.url
import random


class Settings:
    def __init__(
            self,
            language: str = 'en',
            num_results: int = 20,
            merge: bool = True,
            engines: list = None,
            **kwargs
    ):
        """
        For settings only.
        When merge set True, dups will be deleted, and for multi search it will merge results as well.
        engines is a mandatory parameter for multi search - list of engines to use.
        You can set additionall settings in kwargs, like api keys (brave, yandex)
        possible:
            brave_key - brave api key
            ya_key - yandex cloud api key
            ya_fldid - yandex cloud folder id 
        """
        extra_keys = [
            'brave_key', 'ya_key', 'ya_fldid'
        ]
        # TODO: add check for possible lang codes # languages = ['en', ...]
        for k in kwargs:
            if k not in extra_keys:
                raise KeyError("Unexpected key given:", k)

        self.language = language
        self.num_results = num_results
        self.del_dups = merge  # for single search
        self.merge = merge  # for multi search
        self.engines = engines
        self.extra = kwargs


class SearchResult:
    def __init__(self, title, description, url):
        self.title = title
        self.description = description
        self.url = url


def random_agent():
    return random.choice([
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:97.0) Gecko/20100101 Firefox/97.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6; en-US) Gecko/20100101 Firefox/45.4',
        'Mozilla/5.0 (compatible; MSIE 8.0; Windows; U; Windows NT 6.0; x64 Trident/4.0)'
        'Mozilla/5.0 (Linux i563 x86_64; en-US) Gecko/20100101 Firefox/55.0',
        'Mozilla/5.0 (Linux i573 x86_64; en-US) Gecko/20100101 Firefox/56.9',
        'Mozilla/5.0 (Linux; Linux x86_64) Gecko/20100101 Firefox/67.2',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_2) Gecko/20130401 Firefox/68.1',
        'Mozilla/5.0 (U; Linux x86_64; en-US) Gecko/20100101 Firefox/69.5',
        'Mozilla/5.0 (Macintosh; U; Intel Mac OS X 9_3_8; en-US) Gecko/20100101 Firefox/68.8',

    ])


def normalise_url(url):
    url = str(url)
    # Remove the scheme (http(s)://)
    url = re.sub(r'^https?://', '', url)
    # Remove the "www" if it exists
    url = re.sub(r'^www\.', '', url)
    # add slush to the end
    url = (url + "/") if not url.endswith("/") else url
    # decode if needed 
    url = urllib.parse.unquote(url)
    # Canonicalise url
    url = w3lib.url.canonicalize_url(url)
    return url


def merge_results(res_lists: list):
    """
        Merge results from diffirent engines by unique url
    """
    merged = []
    uniq_urls = []

    for lst in res_lists:
        # print("Engine results num =", len(lst))
        for item in lst:
            if item.url not in uniq_urls:
                merged.append(item)
                uniq_urls.append(item.url)

    # print("Total num =", len(merged))
    return merged



from anywebsearch.tools import SearchResult, Settings, normalise_url, merge_results
from brave import Brave
import time


def brave_search(query, lang: str = None, reverse=None, offset=0, settings: Settings = Settings()):
    # Brave search API
    if lang is None:
        lang = settings.language
    try:
        brave_key = str(settings.extra.get('brave_key'))
    except KeyError:
        raise KeyError("Specify the brave_key in the settings class")
    try:
        brave = Brave(brave_key)
        sr = brave.search(
            q=query,
            count=settings.num_results,
            offset=offset if reverse else 0,
            search_lang=lang,
            result_filter='web',
            text_decorations=False,
            country=lang
        )
    except AttributeError as e:
        raise RuntimeError("Bad response! You probably provided an invalid brave_key", e)
    sr = [*sr][3][1].results  # unpack result -> WEB results field -> skip type (0) -> extract results
    sr = [
        SearchResult(result.title, result.description, normalise_url(result.url))
        for result in sr
    ]
    if not reverse and len(sr) < settings.num_results:
        i = 0
        while len(sr) < settings.num_results:
            i += 1
            sr = [*sr, *brave_search(query=query, lang=lang, reverse=sr, offset=i, settings=settings)]
            if settings.del_dups is True:
                sr = merge_results([sr])  # don't merge anything, just del dups
        sr = sr[:settings.num_results]
    return sr



from .tools import merge_results, Settings

from .engines import google_search, qwant_search, duck_search, brave_search, yandex_search

possible_engines = {
    'google': google_search,
    'qwant': qwant_search,
    'duck': duck_search,  # duck is DuckDuckGo (bing engine)
    'brave': brave_search,
    'yandex': yandex_search,
}


def multi_search(query, settings: Settings = Settings()):
    for e in settings.engines:
        if e not in possible_engines.keys():
            raise ValueError(f"Invalid search engine name: {e}")

    results_lists = []
    for e in settings.engines:
        results_lists.append(possible_engines[e](query=query, lang=settings.language, settings=settings))
    if settings.merge is True:
        return merge_results(results_lists)
    return results_lists



import googlesearch as gs

from anywebsearch.tools import SearchResult, Settings, normalise_url, merge_results


def google_search(query, lang: str = None, settings: Settings = Settings()):
    # Google search
    if lang is None:
        lang = settings.language
    sr = gs.search(
        query,
        advanced=True,
        num_results=(
            settings.num_results + int(settings.num_results / 4)
            if settings.del_dups is True
            else settings.num_results
        ),
        # +10 in case of duplicates, and the excess will be trimmed
        lang=lang
    )
    sr = [
        SearchResult(result.title, result.description, normalise_url(result.url))
        for result in sr
    ]
    if settings.del_dups is True:
        sr = merge_results([sr])  # don't merge anything, just del dups
    sr = sr[:settings.num_results]
    return sr



from duckduckgo_search import DDGS

from anywebsearch.tools import SearchResult, Settings, normalise_url, merge_results


def duck_search(query, lang: str = None, settings: Settings = Settings()):
    # ddg search API call
    lang_codes = {
        "AR": "xa-ar",
        "AU": "au-en",
        "AT": "at-de",
        "BE": "be-fr",
        "BR": "br-pt",
        "CA": "ca-en",
        "CL": "cl-es",
        "DK": "dk-da",
        "FI": "fi-fi",
        "FR": "fr-fr",
        "DE": "de-de",
        "HK": "hk-tzh",
        "IN": "in-en",
        "ID": "id-id",
        "IT": "it-it",
        "JP": "jp-jp",
        "KR": "kr-kr",
        "MY": "my-ms",
        "MX": "mx-es",
        "NL": "nl-nl",
        "NZ": "nz-en",
        "NO": "no-no",
        "CN": "cn-zh",
        "PL": "pl-pl",
        "PT": "pt-pt",
        "PH": "ph-en",
        "RU": "ru-ru",
        "SA": "za-en",
        "ZA": None,
        "ES": "es-es",
        "SE": "se-sv",
        "CH": "ch-de",
        "TW": "tw-tzh",
        "TR": "tr-tr",
        "GB": "uk-en",
        "US": "us-en",
        "ALL": "wt-wt",
    }
    if lang is None:
        lang = settings.language
    lang = lang_codes[lang.upper()] if lang.upper() in lang_codes.keys() else "us-en"
    sr = DDGS().text(
        query,
        region=lang,
        max_results=(
            settings.num_results + int(settings.num_results / 4)
            if settings.del_dups is True
            else settings.num_results
        )
        # +25% in case of duplicates, and the excess will be trimmed
    )
    sr = [
        SearchResult(result['title'], result['body'], normalise_url(result['href']))
        for result in sr
    ]
    if settings.del_dups is True:
        sr = merge_results([sr])  # don't merge anything, just del dups
    sr = sr[:settings.num_results]
    return sr



from anywebsearch.tools import SearchResult, Settings, normalise_url, merge_results, random_agent

import xmltodict
import requests


#   _      __  ____    ___    __ __        ____   _  __        ___    ___   ____   _____   ___    ____   ____   ____
#  | | /| / / / __ \  / _ \  / //_/       /  _/  / |/ /       / _ \  / _ \ / __ \ / ___/  / _ \  / __/  / __/  / __/
#  | |/ |/ / / /_/ / / , _/ / ,<         _/ /   /    /       / ___/ / , _// /_/ // (_ /  / , _/ / _/   _\ \   _\ \  
#  |__/|__/  \____/ /_/|_| /_/|_|       /___/  /_/|_/       /_/    /_/|_| \____/ \___/  /_/|_| /___/  /___/  /___/  
#                                                                                                                   

def yandex_search(query, lang: str = None, reverse=None, offset=0, settings: Settings = Settings()):
    if lang is None:
        lang = settings.language
        if lang != "ru":
            lang = "en"

    url = f"https://yandex.com/search/xml?folderid={settings.extra.get('ya_fldid')}&filter=moderate&lr=225&l10n={lang}"
    payload = """<?xml version="1.0" encoding="UTF-8"?>
    <request>
        <query>
            {query}
        </query>
        <groupings>
            <groupby attr="" mode="flat" groups-on-page="10" docs-in-group="1" />
        </groupings>
        <maxpassages>5</maxpassages>
        <page>{offset}</page>
    </request>
    """.format(query=query, offset=offset)
    print(payload)
    headers = {
        'Content-Type': 'application/xml',
        'User-Agent': random_agent(),
        'Authorization': f'Api-Key {settings.extra.get("ya_key")}'
    }
    response = requests.request("POST", url, data=payload, headers=headers, timeout=8)
    response.raise_for_status()
    sr = xmltodict.parse(response.content)

    try:
        sr = sr['yandexsearch']['response']['results']['grouping']['group']
    except KeyError as e:
        if sr['yandexsearch']['response']['error']['@code'] == '31':
            raise RuntimeError("Incorrect yandex API key or folder id")
        raise RuntimeError(e)

    sr = [group['doc'] for group in sr]
    # TODO
    sr = [
        SearchResult(
            result['title']['#text'] if type(result['title']) is dict else result['title'],
            result['passages']['passage']['#text'] if 'passages' in result and type(
                result['passages']['passage']) is dict else "no text",
            normalise_url(result['url'])
        )
        for result in sr
        if (not all(v is None for v in result))
    ]

    if not reverse and len(sr) < settings.num_results:
        i = 0
        while len(sr) < settings.num_results:
            i += 1
            sr = [*sr, *yandex_search(query, lang, reverse=sr, offset=i, settings=settings)]
        sr = sr[:settings.num_results]

    print(sr)



import requests

from anywebsearch.tools import SearchResult, Settings, normalise_url, random_agent, merge_results


def qwant_search(query, lang: str = None, reverse=None, offset=0, settings: Settings = Settings()):
    lang_codes = {
        "AR": None,
        "AU": "en_au",
        "AT": "de_at",
        "BE": "fr_be",
        "BR": "br_fr",
        "CA": "en_ca",
        "CL": "es_cl",
        "DK": "da_dk",
        "FI": "fi_fi",
        "FR": "fr_fr",
        "DE": "de_de",
        "HK": "zh_hk",
        "IN": None,
        "ID": None,
        "IT": "it_it",
        "JP": None,
        "KR": "ko_kr",
        "MY": "en_my",
        "MX": "es_mx",
        "NL": "nl_nl",
        "NZ": "en_nz",
        "NO": "nb_no",
        "CN": "zh_cn",
        "PL": "pl_pl",
        "PT": "pt_pt",
        "PH": None,
        "RU": None,
        "SA": None,
        "ZA": None,
        "ES": "es_es",
        "SE": "sv_se",
        "CH": "de_ch",
        "TW": "zh_tw",
        "TR": None,
        "GB": "en_gb",
        "US": "en_us",
        "ALL": None,
    }
    headers = {'User-Agent': random_agent()}
    if lang is None:
        lang = settings.language
    rlang = lang
    lang = (
        lang_codes[lang.upper()]
        if lang.upper() in lang_codes.keys() and lang_codes[lang.upper()]
        else 'ALL'
    )
    # print("Qwant used lang", lang)
    params = {
        'q': query,
        'offset': offset if reverse else 0,
        # &count is always 10
    }
    if lang != "ALL":
        params['locale'] = lang
    url = 'https://api.qwant.com/v3/search/web'  # ?q=elon%20musk&locale=de_de&offset=0
    raw_sr = requests.get(url, params=params, headers=headers)
    raw_sr.raise_for_status()
    sr = raw_sr.json()  # convert to json
    sr = sr['data']['result']['items']['mainline']  # extract results field from this complex json
    sr = [item for item in sr if item['type'] == 'web']  # extract only web results (there are also videos, ads, images)
    restructed_sr = []
    # There might be a few web blocks with items
    for block in sr:
        for item in block['items']:
            restructed_sr.append(item)
    sr = [
        SearchResult(result['title'], result['desc'], normalise_url(result['url']))
        for result in restructed_sr
    ]
    if not reverse and len(sr) < settings.num_results:
        i = 0
        while len(sr) < settings.num_results:
            i += 10
            sr = [*sr, *qwant_search(query, rlang, reverse=sr, offset=i, settings=settings)]
            if settings.del_dups is True:
                sr = merge_results([sr])  # don't merge anything, just del dups
        sr = sr[:settings.num_results]
    return sr
